// DGen v1.02+
// Megadrive C++ module

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "md.h"

// This is the 'static' StarScream/MZ80 multitasker
// which detects which megadrive is active (by which star_mz80_on() has been called
// and forwards to the appropriate misc_read/writebyte/word/z80 function
// of the appropriate instance (grrrr - bloody C++)

static md* which=0;

int md::star_mz80_on()
{
  s68000SetContext(&cpu);
  mz80SetContext(&z80);
  which=this;
  return 0;
}
int md::star_mz80_off()
{
  which=0;
  s68000GetContext(&cpu);
  mz80GetContext(&z80);
  return 0;
}

unsigned root_readbyte(unsigned a)
{ if (which) return which->misc_readbyte(a); else return 0x00; }
unsigned root_readword(unsigned a)
{ if (which) return which->misc_readword(a); else return 0x0000; }
void root_writebyte(unsigned a,unsigned d)
{ if (which) which->misc_writebyte(a,d); }
void root_writeword(unsigned a,unsigned d)
{ if (which) which->misc_writeword(a,d); }

UINT8 root_z80_read(UINT32 a,struct MemoryReadByte *huh)
{
  (void)(huh);
  if (which) return which->z80_read(a); else return 0x00;
}

void root_z80_write(UINT32 a,UINT8 d,struct MemoryWriteByte *huh)
{
  (void)(huh);
  if (which) which->z80_write(a,d);
}

extern FILE *debug_log;
static int cyc_total=0;

int md::run_to_odo(int odo_to)
{
// THIS FUNCTION IS REALLY A MACRO - it expects star_mz80_on() to have been
// called
  int odo_start=0;
  odo_start=s68000readOdometer();
  s68000exec(odo_to-odo_start);

  if (z80_online)
  {
    int start=0,wanted=0;
    mz80GetElapsedTicks(1);
    wanted=(odo_to-odo_start)/2;
    wanted-=z80_extra_cycles;
    cyc_total+=wanted;
    mz80exec(wanted);
    z80_extra_cycles=(mz80GetElapsedTicks(1)-wanted);
  }
  return 0;
}

extern "C" int mega_dacout=0,mega_dacen=0;
int md::one_frame_hints(struct bmap *bm,REAL_FMSAM *sndbuf[2],int sndlen)
{
//  int vblank_len=(int)1000000*767/100/60;
  int vblank_len=53693100L/7/60;
  int odo_end_of_ras=0,odo_mid_of_ras=0;
  int ras=0;
  int vbl_in=0;
  int hint_countdown=0;

  star_mz80_on(); // VERY IMPORTANT - Must call before using stars/mz80!!
  cyc_total=0;

  s68000tripOdometer();   // odo=0 marks start of frame

  odo_frame_max=vblank_len;

  for (vbl_in=0;vbl_in<2;vbl_in++)
  {
    int ras_st=0,ras_en=0;
    // Genecyst style
    if (vbl_in==0)
    {
      // End of VBLANK
      coo4=0x02; coo5=0x24;
      ras_st=0; ras_en=224;
    }
    else
    {
      // Start of VBLANK
      coo4=0x02; coo5=0xa8;
      ras_st=224; ras_en=224+38;
    }

    for (ras=ras_st;ras<ras_en;ras++)
    {
      if (ras==(224+38)/2) if (bm!=NULL) //smack in the middle
      {
        // Render the graphics when it is most likely to look right!
        // (smack in the middle of the screen!)
        struct dgen_sinfo si;
        vdp.get_screen_info(&si);
        draw_md_graphics(bm,NULL,&si);
      }

      if (ras==224+10)
      {
        if (vdp.reg[1]&0x20)
        {
          s68000interrupt(6);
        }
        mz80int(0x38);
      }

      odo_end_of_ras=(int)(ras+1)*vblank_len/(224+38);
      odo_mid_of_ras=(int)(ras+0)*vblank_len/(224+38);
      odo_mid_of_ras=(odo_mid_of_ras+odo_end_of_ras)/2;

      if (hint_countdown<=0)
      {
        if (vdp.reg[0]&0x10) s68000interrupt(4);
        hint_countdown=(vdp.reg[0xa]+1)*2;
        // *2 seems to correct music
      }
      coo5|=0x04;
      run_to_odo(odo_mid_of_ras);
      coo5&=0xfb;
      run_to_odo(odo_end_of_ras);

      hint_countdown--;

      if (sndbuf!=NULL)
      {
        REAL_FMSAM *trabuf[2];

        int sndfrom,sndto;
        sndfrom=(ras+0)*sndlen/(224+38);
        sndto  =(ras+1)*sndlen/(224+38);
        trabuf[0]=sndbuf[0]+sndfrom;
        trabuf[1]=sndbuf[1]+sndfrom;
        YM2612UpdateOne(0,(void **)trabuf,sndto-sndfrom);

        signed short ssdac;
        ssdac=0;
        if (mega_dacout)
        {
          ssdac=(mega_dacout<<6);
          ssdac-=0x2000;
        }
        /*
        {
          static signed short test[500]={0};
          SN76496Update_16(0,test,500);
          int i;
          for (i=0;i<500;i++) if (test[i]!=0) exit(0);

        }
        */

        int i;
        for (i=0;i<(sndto-sndfrom);i++)
        {
          trabuf[0][i]>>=1;
          trabuf[1][i]>>=1;

          trabuf[0][i]+=ssdac;
          trabuf[1][i]+=ssdac;
        }

      }
    }
  }

  star_mz80_off(); // VERY IMPORTANT - Must call after using stars/mz80!!


  return 0;
}

int md::one_frame()
{
  int vblank_start=(int)1000000*767/100*5/60/6;
  int vblank_end=  (int)1000000*767/100*6/60/6;
  star_mz80_on(); // VERY IMPORTANT - Must call before using stars/mz80!!

  s68000tripOdometer(); // odo=0 marks start of frame

  odo_frame_max=vblank_end;
  // For coo8 values, tell mem.cc the odo at the end

  // Genecyst style
  coo4=0x02;
  coo5=0x24;

  run_to_odo(vblank_start);

  coo4=0x02;
  coo5=0xa8;
  if (vdp.reg[1]&0x20) s68000interrupt(6);

  run_to_odo(vblank_end);

  star_mz80_off(); // VERY IMPORTANT - Must call after using stars/mz80!!
  return 0;
}


static struct MemoryReadByte mem_read[]=
{
  {0x2000,0xffff,root_z80_read},
  {(UINT32) -1,(UINT32) -1,NULL}
};
static struct MemoryWriteByte mem_write[]=
{
  {0x2000,0xffff,root_z80_write},
  {(UINT32) -1,(UINT32) -1,NULL}
};
static struct z80PortRead  io_read[] ={{(UINT16) -1,(UINT16) -1,NULL}};
static struct z80PortWrite io_write[]={{(UINT16) -1,(UINT16) -1,NULL}};

int md::memory_map()
{
  int i=0,j=0;
  int rommax=romlen;

  if (rommax>0xa00000) rommax=0xa00000;
  if (rommax<0) rommax=0;

// FETCH: Set up 2 or 3 FETCH sections
  i=0;
  if (rommax>0)
  { fetch[i].lowaddr=0x000000; fetch[i].highaddr=rommax-1; fetch[i].off=(unsigned)rom-0x000000; i++; }
  fetch[i].lowaddr=0xff0000; fetch[i].highaddr=0xffffff; fetch[i].off=(unsigned)ram-  0xff0000; i++;
// Testing
  fetch[i].lowaddr=0xffff0000; fetch[i].highaddr=0xffffffff; fetch[i].off=(unsigned)ram-0xffff0000; i++;
// Testing 2
  fetch[i].lowaddr=0xff000000; fetch[i].highaddr=0xff000000+rommax-1; fetch[i].off=(unsigned)rom-0xff000000; i++;
  fetch[i].lowaddr=fetch[i].highaddr=0xffffffff; fetch[i].off=0; i++;

  i=0; j=0;

#if 0
// Simple version ***************
  readbyte[i].lowaddr=   readword[i].lowaddr=
  writebyte[j].lowaddr=  writeword[j].lowaddr=   0;
  readbyte[i].highaddr=  readword[i].highaddr=
  writebyte[j].highaddr= writeword[j].highaddr=  0xffffffff;

  readbyte[i].memorycall=(void *) root_readbyte;
  readword[i].memorycall=(void *) root_readword;
  writebyte[j].memorycall=(void *)root_writebyte;
  writeword[j].memorycall=(void *)root_writeword;

  readbyte[i].userdata=  readword[i].userdata=
  writebyte[j].userdata= writeword[j].userdata=  NULL;
  i++; j++;
// Simple version end ***************

#else
// Faster version ***************
// IO: Set up 3/4 read sections, and 2/3 write sections
  if (rommax>0)
  {
// Cartridge ROM memory (read only)
    readbyte[i].lowaddr=   readword[i].lowaddr=   0x000000;
    readbyte[i].highaddr=  readword[i].highaddr=  rommax-1;
    readbyte[i].memorycall=readword[i].memorycall=NULL;
    readbyte[i].userdata=  readword[i].userdata=  rom;
    i++;

// misc memory (e.g. aoo and coo) through root_rw
    readbyte[i].lowaddr=   readword[i].lowaddr=
    writebyte[j].lowaddr=  writeword[j].lowaddr=   rommax;
  }
  else
    readbyte[i].lowaddr=   readword[i].lowaddr=
    writebyte[j].lowaddr=  writeword[j].lowaddr=   0;

  readbyte[i].highaddr=  readword[i].highaddr=
  writebyte[j].highaddr= writeword[j].highaddr=  0xfeffff;

  readbyte[i].memorycall=(void *) root_readbyte;
  readword[i].memorycall=(void *) root_readword;
  writebyte[j].memorycall=(void *)root_writebyte;
  writeword[j].memorycall=(void *)root_writeword;

  readbyte[i].userdata=  readword[i].userdata=
  writebyte[j].userdata= writeword[j].userdata=  NULL;
  i++; j++;

// scratch RAM memory
  readbyte[i].lowaddr =   readword[i].lowaddr =
  writebyte[j].lowaddr =  writeword[j].lowaddr =   0xff0000;
  readbyte[i].highaddr=   readword[i].highaddr=
  writebyte[j].highaddr=  writeword[j].highaddr=   0xffffff;
  readbyte[i].memorycall= readword[i].memorycall=
  writebyte[j].memorycall=writeword[j].memorycall= NULL;
  readbyte[i].userdata=  readword[i].userdata =
  writebyte[j].userdata= writeword[j].userdata =   ram;
  i++; j++;
// Faster version end ***************
#endif

// The end
   readbyte[i].lowaddr  =   readword[i].lowaddr  =
  writebyte[j].lowaddr  =  writeword[j].lowaddr  =
   readbyte[i].highaddr =   readword[i].highaddr =
  writebyte[j].highaddr =  writeword[j].highaddr = 0xffffffff;

   readbyte[i].memorycall=  readword[i].memorycall=
  writebyte[j].memorycall= writeword[j].memorycall=
   readbyte[i].userdata=    readword[i].userdata =
  writebyte[j].userdata=   writeword[j].userdata = NULL;
  i++; j++;

  return 0;
}


int md::reset()
{
  star_mz80_on();
  s68000reset();
  mz80reset();
  z80_bank68k=0x000; // 9 bits
  z80_online=0;
  z80_extra_cycles=0;
  star_mz80_off();
  return 0;
}

int md::point_mz80_stuff()
{
  z80.z80Base=z80ram;
  z80.z80MemRead=mem_read;
  z80.z80MemWrite=mem_write;
  z80.z80IoRead=io_read;
  z80.z80IoWrite=io_write;
  return 0;
}

int md::z80_init()
{
  // Set up the z80
  star_mz80_on();
  mz80reset();
  // Modify the default context
  mz80GetContext(&z80);
  point_mz80_stuff();
  mz80SetContext(&z80);
  mz80reset();
  star_mz80_off();
  return 0;
}

int md::point_star_stuff()
{
  cpu.memoryfetch =     fetch;
  cpu.memoryreadbyte =  readbyte;
  cpu.memoryreadword =  readword;
  cpu.memorywritebyte = writebyte;
  cpu.memorywriteword = writeword;
  return 0;
}


md::md()
{
  romlen=ok=0; 
  mem=rom=ram=z80ram=NULL;
  memset(&cpu,0,sizeof(cpu));
  memset(&z80,0,sizeof(z80));
  fetch=NULL; readbyte=readword=writebyte=writeword=NULL;
  z80_bank68k=z80_online=odo_frame_max=z80_extra_cycles
    =coo_waiting=coo_cmd=aoo3_toggle=aoo5_toggle
    =check_pad=pad=coo4=coo5=0;
  memset(romfilename,0,sizeof(romfilename));

  ok=0;
  if (!vdp.okay()) return;
  vdp.belongs=this;

  coo_waiting=0; coo_cmd=0; check_pad=0;
  pad=0; coo4=0; coo5=0;
  aoo3_toggle=aoo5_toggle=0x00; // zero = natural state?
  romfilename[0]=0;

  odo_frame_max=0;

  //  Format of pad is: __SA____ UDLRBC__

  rom=mem=ram=z80ram=NULL;
  mem=(unsigned char *)malloc(0x20000);
  if (mem==NULL) return;
  memset(mem,0,0x20000);
  ram=   mem+0x00000;
  z80ram=mem+0x10000;

  romlen=0;

  fetch= new S68000MEMORYFETCH[5]; if (!fetch)     return;
  readbyte= new S68000MEMORYIO[4]; if (!readbyte)  return;
  readword= new S68000MEMORYIO[4]; if (!readword)  return;
  writebyte=new S68000MEMORYIO[4]; if (!writebyte) return;
  writeword=new S68000MEMORYIO[4]; if (!writeword) return;

  memory_map();

  point_star_stuff();

  star_mz80_on();  // VERY IMPORTANT - Must call before using stars/mz80!!
  if (s68000init()!=0) { printf ("s68000init failed!\n"); return; }
  star_mz80_off(); // VERY IMPORTANT - Must call after using stars/mz80!!

  z80_init();

  reset(); // reset megadrive

  ok=1;
}


md::~md()
{
  romfilename[0]=0;
  if (rom!=NULL) unplug();

  free(mem);
  rom=mem=ram=z80ram=NULL;

  if (fetch)     delete[] fetch;
  if (readbyte)  delete[] readbyte;
  if (readword)  delete[] readword;
  if (writebyte) delete[] writebyte;
  if (writeword) delete[] writeword;

  ok=0;
}

// Byteswaps memory
int byteswap_memory(unsigned char *start,int len)
{ int i; unsigned char tmp;
  for (i=0;i<len;i+=2)
  { tmp=start[i+0]; start[i+0]=start[i+1]; start[i+1]=tmp; }
  return 0;
}

int md::plug_in(unsigned char *cart,int len)
{
  // Plug in the cartridge specified by the uchar *
  // NB - The megadrive will free() it if unplug() is called, or it exits
  // So it must be a single piece of malloced data
  if (cart==NULL) return 1; if (len<=0) return 1;
  byteswap_memory(cart,len); // for starscream
  romlen=len;
  rom=cart;
  memory_map(); // Update memory map to include this cartridge
  reset(); // Reset megadrive
  return 0;
}

int md::unplug()
{
  if (rom==NULL) return 1; if (romlen<=0) return 1;
  free(rom); romlen=0;
  memory_map(); // Update memory map to include no rom
  return 0;
}

extern "C" int load_rom_into(char *name,unsigned char *into);

int md::load(char *name)
{
 // Convenience function - calls romload.c functions
 unsigned char *temp=NULL; int len=0;
 if (name==NULL) return 1;

 len=load_rom_into(name,NULL);
 if (len<=0) return 1;
 temp=(unsigned char *)malloc(len);
 if (temp==NULL) return 1;
 load_rom_into(name,temp);
 // Register name
 strncpy(romfilename,name,255);
 plug_in(temp,len); // md then deallocates it when it's done
 
 return 0;
}

int md::z80dump()
{
  FILE *hand;
  hand=fopen("dgz80ram","wb");
  if (hand!=NULL)
  { fwrite(z80ram,1,0x10000,hand); fclose(hand); }
  return 0;
}
