// DGen v1.03+

#define VER "1.03"

#ifndef __STARCPU_H__
#include <starcpu.h>
#endif

#ifndef	_MZ80_H_
#include <mz80.h>
#endif

#define BUILD_YM2612
extern "C" {
#include "fm.h"
}

extern "C" {
int SN76496_init(int chip,int clock,int sample_rate,int sample_bits);
void SN76496Write(int chip,int data);
void SN76496Update_16(int chip,void *buffer,int length);
}

typedef signed short REAL_FMSAM;

struct bmap { unsigned char *data; int w,h; int pitch; int bpp; };

// Info passed to render the screen (possibly line by line)
struct dgen_sinfo { unsigned char *vram,*cram,*vsram,*vdp_reg; };

int draw_md_graphics
  (struct bmap *bm,unsigned char passpal[256],struct dgen_sinfo *si);
int get_md_palette(unsigned char pal[256],unsigned char *cram);

class md;
class md_vdp
{
public:
  int ok;
  unsigned char *mem,*vram,*cram,*vsram;
private:
  int dma_len();
  int dma_addr();
  unsigned char dma_mem_read(int addr);
  int putword(unsigned short d);
  int putbyte(unsigned char d);
public:
  int rw_mode,rw_addr,rw_dma;
  md *belongs;
  unsigned char reg[0x20];
  int okay() {return ok;}
  md_vdp();
  ~md_vdp();
  int command(unsigned int cmd);
  unsigned short readword();
  unsigned char readbyte();
  int writeword(unsigned short d);
  int writebyte(unsigned char d);
  int get_screen_info(struct dgen_sinfo *si);
};

class md
{
private:
  int romlen;
public:
  int ok; 
  unsigned char *mem,*rom,*ram,*z80ram;
  md_vdp vdp;
  struct S68000CONTEXT cpu;
  struct mz80context z80;
private:
  S68000MEMORYFETCH *fetch;
  S68000MEMORYIO    *readbyte,*readword,*writebyte,*writeword;
  int memory_map();
  int star_mz80_on();
  int star_mz80_off();

  int z80_bank68k; // 9 bits
  int z80_online;
  int odo_frame_max; // value of odo at end of frame
  int z80_extra_cycles;

  // Note order is (0) Vblank end -------- Vblank Start -- (HIGH)
  // So int6 happens in the middle of the count

  int coo_waiting; // after 04 write which is 0x4xxx or 0x8xxx
  unsigned int coo_cmd; // The complete command
  int aoo3_toggle,aoo5_toggle;
  int check_pad;
  inline unsigned short calculate_coo8();
  int run_to_odo(int odo_to);

  int point_mz80_stuff();
  int point_star_stuff();

public:
  int one_frame();
  int one_frame_hints(struct bmap *bm,REAL_FMSAM *sndbuf[2],int sndlen);

  int pad;
// c000004 bit 1 write fifo empty, bit 0 write fifo full (???)
// c000005 vint happened, (sprover, coll, oddinint)
// invblank, inhblank, dma busy, pal
  unsigned char coo4,coo5;
  int okay() {return ok;}
  md();
  ~md();
  int plug_in(unsigned char *cart,int len);
  int unplug();
  int load(char *name);

  int reset();

  unsigned misc_readbyte(unsigned a);
  void misc_writebyte(unsigned a,unsigned d);
  unsigned misc_readword(unsigned a);
  void misc_writeword(unsigned a,unsigned d);

  int z80_init();
  void z80_write(unsigned int a,unsigned char v);
  unsigned char z80_read(unsigned int a);

  S68000CONTEXT& cpu_context() {return cpu;}
  mz80context&   z80_context() {return z80;}
  int load(FILE *hand,char *desc);
  int save(FILE *hand,char *desc);
  int import_gst(FILE *hand);
  int export_gst(FILE *hand);

  char romfilename[256]; // If known

  int get_screen_info(struct dgen_sinfo *si)
  { return vdp.get_screen_info(si); }
  int z80dump();
};
