// DGen v1.09+

#include <stdio.h>
#include <stdlib.h>
#include "md.h"

static char tmp[1024]="";

/*
   PR | P1 | P0 | VF | HF | XXXXXXXXXX
   PR = Priority bit. If bit is set tile is shown in front of sprites.
   P1 - P0 = Sets palette of tile. (0 - 3)
   VF - Vertically flips tile.
   HF - Horizontally flips tile.
   The last ten bits are the tile number (0 - 2047)
*/

static unsigned int highpal[64]={0};
static unsigned char *bmp_data=NULL;
static int bmp_depth=8,lpitch=240+16;
static int bmp_inc=1; // Amount to increase to get to next pixel
static unsigned char *vram=NULL,*vsram=NULL,*cram=NULL;
static unsigned char *reg=NULL;
extern FILE *debug_log;

int graph_y_top=0;
int graph_y_bottom=224;

static inline int depth_to_bytes(int depth)
{
       if (depth<=8)  return 1;
  else if (depth<=16) return 2;
  else if (depth<=24) return 3;
  else                return 4;
}


#ifndef ASM_TILES
// C Version of draw_tile
static int draw_tile(int ox,int oy,unsigned tile)
{
  int tile_off,pal_off;
  int p1,p0,fx,fy,x,y;

  // shift by eight for messy border
  ox+=8; oy+=8;
  if (ox<0) return 1;
  if (ox>=(320+8)) return 1;
  if (oy<0) return 1;
  if (oy>=(224+8)) return 1;


  if ((reg[12]&3)==3)        // interlace mode
    tile_off=(tile&2047)<<6;
  else
    tile_off=(tile&2047)<<5;

  p0=(tile&0x2000)?1:0;
  p1=(tile&0x4000)?1:0;
  pal_off=(p1*2+p0)*16;

  fy=(tile&0x1000)?1:0;
  fx=(tile&0x0800)?1:0;

  for (y=0;y<8;y++)
  {
    unsigned char b;
    int col; int tx;
    tx=x;
    for (x=0;x<8;x+=2)
    {
      b =vram[(tile_off++)&0xffff];

//      col=pal_off+(b>>4);
//      if (b>>4) _putpixel(bmp_data,fx?(ox+7-x+0):(ox+x+0),fy?(oy+7-y):(oy+y),highpal[col&63]);
//      col=pal_off+(b&15);
//      if (b&15) _putpixel(bmp_data,fx?(ox+7-x-1):(ox+x+1),fy?(oy+7-y):(oy+y),highpal[col&63]);


      if (bmp_depth<=8)
      {
        col=pal_off+(b>>4);
        if (b>>4)
          *((unsigned char  *)(bmp_data+(oy+(fy?(7-y):y))*lpitch+1*(ox+(fx?(7-x+0):(x+0)))))=col&63;
  
        col=pal_off+(b&15);
        if (b&15)
          *((unsigned char  *)(bmp_data+(oy+(fy?(7-y):y))*lpitch+1*(ox+(fx?(7-x-1):(x+1)))))=col&63;
      }
      else if (bmp_depth<=16)
      {
        col=pal_off+(b>>4);
        if (b>>4)
          *((unsigned short *)(bmp_data+(oy+(fy?(7-y):y))*lpitch+2*(ox+(fx?(7-x+0):(x+0)))))=highpal[col&63];
  
        col=pal_off+(b&15);
        if (b&15)
          *((unsigned short *)(bmp_data+(oy+(fy?(7-y):y))*lpitch+2*(ox+(fx?(7-x-1):(x+1)))))=highpal[col&63];
      }
      else if (bmp_depth<=24)
      {
        int clr,cr,cg,cb; unsigned char *pix;
        col=pal_off+(b>>4); clr=highpal[col&63];
        cr=clr>>16; cg=clr>>8; cb=clr; cr&=0xff; cg&=0xff; cb&=0xff;
        if (b>>4)
        {
          pix=(bmp_data+(oy+(fy?(7-y):y))*lpitch+3*(ox+(fx?(7-x+0):(x+0))));
          pix[0]=cb; pix[1]=cg; pix[2]=cr;
        }

        col=pal_off+(b&15); clr=highpal[col&63];
        cr=clr>>16; cg=clr>>8; cb=clr; cr&=0xff; cg&=0xff; cb&=0xff;
        if (b&15)
        {
          pix=(bmp_data+(oy+(fy?(7-y):y))*lpitch+3*(ox+(fx?(7-x-1):(x+1))));
          pix[0]=cb; pix[1]=cg; pix[2]=cr;
        }
      }
      else if (bmp_depth<=32)
      {
        col=pal_off+(b>>4);
        if (b>>4)
          *((unsigned int   *)(bmp_data+(oy+(fy?(7-y):y))*lpitch+4*(ox+(fx?(7-x+0):(x+0)))))=highpal[col&63];
  
        col=pal_off+(b&15);
        if (b&15)
          *((unsigned int   *)(bmp_data+(oy+(fy?(7-y):y))*lpitch+4*(ox+(fx?(7-x-1):(x+1)))))=highpal[col&63];
      }
    }
    if ((reg[12]&3)==3) tile_off+=4;
  }

  return 0;
}
#endif


#ifdef ASM_TILES
#include "gra.h"

// Assembler version of draw_tile
static int draw_tile(int ox,int oy,unsigned tile)
{
  int tile_off,pal_off;
  int fx,fy;
  unsigned char *dest,*tilep;
  if (ox<-8)   return 1;
  if (ox>=320) return 1;
  if (oy<graph_y_top-8) return 1;
  if (oy>=graph_y_bottom) return 1;
  // shift by eight for messy border
  ox+=8; oy+=8;

  if ((reg[12]&3)==3)        // interlace mode
    tile_off=(tile&2047)<<6;
  else
    tile_off=(tile&2047)<<5;

  pal_off=(tile&0x6000)>>9;

  fy=(tile&0x1000)?1:0;
  fx=(tile&0x0800)?1:0;

//  b =vram[(tile_off++)&0xffff];

// e.g. tile_off: 00 12 23 32
  tile_off&=0xffff;
  if ((tile_off+4*8)>=0x10000) return 1;
  tilep=vram+tile_off;
  
// Okay bx runs from 0-63 where (00yyyxxx) is the pixel within the tile
  dest=bmp_data+(oy*lpitch)+ox*bmp_inc;

  // tf_tab contains all the variants
  {
    int fn_num=0;
    if (fx) fn_num|=1; if (fy) fn_num|=2;
         if (bmp_depth<=8)         ;
    else if (bmp_depth<=16) fn_num|=4;
    else if (bmp_depth<=24) fn_num|=8;
    else                    fn_num|=12;
    if ((reg[12]&3)==3) fn_num|=16; // interlace mode

    (tf_tab[fn_num])(dest,lpitch,tilep,highpal+pal_off);
    return 0;
  }
  if (bmp_depth>=32)
  {
    if ( !fx  && !fy  ) draw_tile_32      (dest,lpitch,tilep,highpal+pal_off);
    if (  fx  && !fy  ) draw_tile_32_hf   (dest,lpitch,tilep,highpal+pal_off);
    if ( !fx  &&  fy  ) draw_tile_32_vf   (dest,lpitch,tilep,highpal+pal_off);
    if (  fx  &&  fy  ) draw_tile_32_hf_vf(dest,lpitch,tilep,highpal+pal_off);
  }
  else if (bmp_depth>=24)
  {
    if ( !fx  && !fy  ) draw_tile_24      (dest,lpitch,tilep,highpal+pal_off);
    if (  fx  && !fy  ) draw_tile_24_hf   (dest,lpitch,tilep,highpal+pal_off);
    if ( !fx  &&  fy  ) draw_tile_24_vf   (dest,lpitch,tilep,highpal+pal_off);
    if (  fx  &&  fy  ) draw_tile_24_hf_vf(dest,lpitch,tilep,highpal+pal_off);
  }
  else if (bmp_depth>=15)
  {
    if ( !fx  && !fy  ) draw_tile_16      (dest,lpitch,tilep,highpal+pal_off);
    if (  fx  && !fy  ) draw_tile_16_hf   (dest,lpitch,tilep,highpal+pal_off);
    if ( !fx  &&  fy  ) draw_tile_16_vf   (dest,lpitch,tilep,highpal+pal_off);
    if (  fx  &&  fy  ) draw_tile_16_hf_vf(dest,lpitch,tilep,highpal+pal_off);
  }
  else if (bmp_depth>=8)
  {
    if ( !fx  && !fy  ) draw_tile_8      (dest,lpitch,tilep,highpal+pal_off);
    if (  fx  && !fy  ) draw_tile_8_hf   (dest,lpitch,tilep,highpal+pal_off);
    if ( !fx  &&  fy  ) draw_tile_8_vf   (dest,lpitch,tilep,highpal+pal_off);
    if (  fx  &&  fy  ) draw_tile_8_hf_vf(dest,lpitch,tilep,highpal+pal_off);
  }

  return 0;
}
#endif

/*
Register #11
 
 Bit 2   : Vertical Scroll Control 1: An offset per 16 pixels
                           0: A Screen offset
 Bit 1-0 : Horizontal Scroll Control
            00 : A Screen offset
            01 : Not Used
            10 : An offset per cellule
            11 : An offser per line
 
Register #13
 
 Address of base for Horizontal Scrolls/$400         (B800)
 
*/

static int draw_plane(int ab,int high)
{
  int psx,psy; // playfield size
  int x=0,y=0,maxx=40,fixx=0;
  int hscroll_base=0,hscrolloff;
  static int playinfo[4]={0x20,0x40,0x40,0x80};

  psx=reg[0x10]; psy=psx>>4; psx&=3; psy&=3;
  // 00=32 cell 01=64 cell, 11=128 cell
  psx=playinfo[psx];
  psy=playinfo[psy];

  hscroll_base=(reg[13]<<10)+ab*2;

//  for (y=-1;y<28;y++)
  for (y=(graph_y_top/8)-1;y<(graph_y_bottom/8);y++)
  {
    int pl;
    unsigned short xoff=0,yoff=0;
    int add=0;
    int window_all=0;

    if (ab<2)
    {
      yoff= vsram[(ab*2+0)&0x7f];     yoff<<=8; yoff|=vsram[(ab*2+1)&0x7f];

      if ((reg[12]&3)==3)
      {
        yoff/=2; // interlace mode
      }

      if ((reg[0xb]&3)<=1) hscrolloff=0; else hscrolloff=y*4*8;
      if (hscrolloff<0) hscrolloff=0;
      xoff= vram[(hscroll_base+hscrolloff+0)&0xffff]; xoff<<=8;
      xoff|=vram[(hscroll_base+hscrolloff+1)&0xffff];
    }

    if (ab==0)      pl=(reg[0x02]<<10);
    else if (ab==1) pl=(reg[0x04]<<13);
    else
    {
      // Window
      pl=(reg[0x03]<<10);

      window_all=0;
      if (reg[0x12]&0x80)
      {
        if (y>=(reg[0x12]&0x1f)) window_all=1;
      }
      else
      {
        if (y<(reg[0x12]&0x1f)) window_all=1;
      }
      psy=64;
      if ((reg[0xc]&1)==0) // 32 horiz
        psx=32; else psx=64;
    }

    yoff=-yoff;
    pl+=((-((yoff)>>3)+y)&(psy-1))*psx*2;
    add=(-(xoff>>3)-1)<<1;
    if ((reg[0xc]&1)==0) { maxx=32; fixx=4; } else { maxx=40; fixx=0; }

    for (x=-1;x<maxx;x++)
    {
      unsigned int tile;
      int ta;

      if ((ab>=2)&&(!window_all))
      {
        // Window
        if (reg[0x11]&0x80)
        {
          if (x<  ((reg[0x11]&0x1f)<<1) ) goto xskip;
        }
        else
        {
          if (x>= ((reg[0x11]&0x1f)<<1) ) goto xskip;
        }
      }

      ta=(pl+(add&((psx-1)<<1)));
      tile=(vram[(ta+0)&0xffff]<<8)+vram[(ta+1)&0xffff];
      if ( ((tile&0x8000)==0x8000) == high )
        draw_tile(((x+fixx)<<3)+(xoff&7),(y<<3)+(yoff&7),tile);
      xskip:
      add+=2;
    }
  
    switch(reg[0xb]&3)
    {
//      case 2: hscroll+=4*8; break;
      // Although it is offset per tile, the offsets are 32 bytes apart
//      case 3: hscroll+=4*8; break;
      // case3 is a fudge - should be more accurate and displace within tiles
    }
  }

  return 0;
}

static int draw_sprites(int high)
{
  static unsigned sprite_order[0x100]={0};
  int spr=0,sprite_count=0,i=0;

  spr=(reg[0x05]<<9);
  {
    int i;
    // Calculate the sprite order (needs to be rendered backwards)
    sprite_count=0;
    sprite_order[0]=0;
    for (i=1;i<=0x100;i++)
    {
      unsigned char next;
      next=vram[(spr+sprite_order[i-1]*8+3)&0xffff];
      if ((next==0)||(i>=0x100)) { sprite_count=i; break; }
      sprite_order[i]=next;
    }
  }

  for (i=sprite_count-1;i>=0;i--)
  {
/*
byte 0-1 : Y-axis (0 to 511)+$80
byte 2   : SIze    (bit 0-1 : height 0-3+1 , Bit 2-3 : width 0-3+1)
byte 3   : Link Data ( next sprite in the list or 0 for end of list)
byte 4-5 : First number tile sprite (Explained below)
bute 6-7 : X-axis (0 to 511)+$80
*/
    int sprno,x,y,xs,ys,tile,tx,ty;
    int fx,fy;

    sprno=sprite_order[i];
    y=(vram[(spr+sprno*8+0)&0xffff]<<8)+vram[(spr+sprno*8+1)&0xffff];
    x=(vram[(spr+sprno*8+6)&0xffff]<<8)+vram[(spr+sprno*8+7)&0xffff];

    x&=0x1ff;
    if ((reg[12]&3)==3)
    {
      // interlace mode
      y&=0x3ff;
      y/=2; // interlace mode
    }
    else 
      y&=0x1ff;

    y-=0x80;
    x-=0x80;

    if ((reg[0xc]&1)==0) x+=4*8;
    ys=vram[(spr+sprno*8+2)&0xffff];
    xs=ys>>2; xs&=3; ys&=3; xs++; ys++;
    tile=vram[(spr+sprno*8+4)&0xffff]; tile<<=8;
    tile+=vram[(spr+sprno*8+5)&0xffff];

    fy=(tile&0x1000)?1:0;
    fx=(tile&0x0800)?1:0;
    for (tx=0;tx<xs;tx++)
    {
      for (ty=0;ty<ys;ty++)
      {
        int dx=x+(fx?xs-tx-1:tx)*8, dy=y+(fy?ys-ty-1:ty)*8;

        if ( ((tile&0x8000)==0x8000) == high )
          draw_tile(dx,dy,tile);
        tile++;
      }
    }
  }
  return 0;
}

// This allows you to get the palette, mess around with it
// (e.g. contrast/brightness)
// and then pass it back
int get_md_palette(unsigned char pal[256],unsigned char *cram)
{
  int c;
  if (pal==NULL) return 1;
  for (c=0;c<64;c++)
  {
    int r,g,b;
    b=(cram[c*2+0]&0x0e)<<4;
    g=(cram[c*2+1]&0xe0);
    r=(cram[c*2+1]&0x0e)<<4;
    pal[c*4+0]=r;
    pal[c*4+1]=g;
    pal[c*4+2]=b;
  }
  return 0;
}

// Must pass a bmp with size 320+16 by 240+16, its depth, and also
// its lPitch
// PASSPAL is now INPUT not output - use get_md_palette first
int draw_md_graphics(struct bmap *bm,unsigned char passpal[256],struct dgen_sinfo *si)
{
  unsigned char ifneedpal[256],*pal;
  bmp_data=bm->data;
  bmp_depth=bm->bpp; lpitch=bm->pitch;
  ::vram=si->vram;
  ::vsram=si->vsram;
  ::cram=si->cram;
  ::reg=si->vdp_reg;
  bmp_inc=1;
       if (bmp_depth>24) bmp_inc=4;
  else if (bmp_depth>16) bmp_inc=3;
  else if (bmp_depth>8)  bmp_inc=2;


  if (graph_y_bottom<graph_y_top) graph_y_bottom=graph_y_top;
  if (graph_y_top<0) graph_y_top=0;
  else if (graph_y_top>224) graph_y_top=224;
  if (graph_y_bottom<0) graph_y_bottom=0;
  else if (graph_y_bottom>224) graph_y_bottom=224;

  if (passpal!=0) pal=passpal;
  else
  {
    get_md_palette(ifneedpal,si->cram);
    pal=ifneedpal;
  }

// Get colors (decoded in above function)
  {
    int c;
    for (c=0;c<64;c++)
    {
      int r,g,b;
      r=pal[c*4+0];
      g=pal[c*4+1];
      b=pal[c*4+2];

      if (bmp_depth<=8) highpal[c]=c;
      else if (bmp_depth<=15) highpal[c]=((r>>3)<<10) + ((g>>3)<<5) + (b>>3);
      else if (bmp_depth<=16) highpal[c]=((r>>3)<<11) + ((g>>2)<<5) + (b>>3);
      else if (bmp_depth<=32) highpal[c]=(r<<16) + (g<<8) + b;
    }
  }

  //clear_to_color(bmp,);
#ifdef ASM_TILES
  {
    unsigned char *start; int bytes; unsigned int colr;
    start=bmp_data+(8+graph_y_top)*lpitch;
    bytes=(graph_y_bottom-graph_y_top)*lpitch;
    if (bmp_depth<=8) colr=reg[7]&63;
    else colr=highpal[reg[7]&63];

// Note - this fast version assumes you can write across the entire bitmap:
// it definitely won't work on a screen bitmap!
    if (bmp_depth<=8)
      clear_to_col_8    (start,bytes/1,colr);
    else if (bmp_depth<=16)
           clear_to_col_16(start,bytes/2,colr);
    else if (bmp_depth<=24)
           clear_to_col_24(start,bytes/3,colr);
    else if (bmp_depth<=32)
           clear_to_col_32(start,bytes/4,colr);
  }
#else
  {
    int x,y;
    if (bmp_depth<=8)
    {
      for (y=8+graph_y_top;y<8+graph_y_bottom;y++)
        for (x=8;x<320+8;x++)
          *((unsigned char *)(bmp_data+y*lpitch+x*1))=reg[7]&63;
    }
    else if (bmp_depth<=16)
    {
      for (y=8+graph_y_top;y<8+graph_y_bottom;y++)
        for (x=8;x<320+8;x++)
          *((unsigned short *)(bmp_data+y*lpitch+x*2))=highpal[reg[7]&63];
    }
    else if (bmp_depth<=24)
    {
      for (y=8+graph_y_top;y<8+graph_y_bottom;y++)
        for (x=8;x<320+8;x++)
        {
          int clr,cr,cg,cb; unsigned char *pix;
          clr=highpal[reg[7]&63];
          cr=clr>>16; cg=clr>>8; cb=clr; cr&=0xff; cg&=0xff; cb&=0xff;
          pix=(bmp_data+y*lpitch+x*3);
          pix[0]=cb; pix[1]=cg; pix[2]=cr;
        }
    }
    else if (bmp_depth<=32)
    {
      for (y=8+graph_y_top;y<8+graph_y_bottom;y++)
        for (x=8;x<320+8;x++)
          *((unsigned int *)(bmp_data+y*lpitch+x*4))=highpal[reg[7]&63];
    }
  }
#endif


// draw a and b low/high and sprites in the right order
  draw_plane(1,0); 
  draw_plane(0,0); 
  draw_plane(2,0);  // window
  draw_sprites(0); 
  draw_plane(1,1);
  draw_plane(0,1); 
  draw_plane(2,1);  // window
  draw_sprites(1); 

#ifdef ASM_TILES
  if ((reg[0xc]&1)==0)
  {
    int y,colr;
    // Remove the messy edges of the thinner mode
    unsigned char *lne=bmp_data+lpitch*8;

    if (bmp_depth<=8) colr=reg[7]&63;
    else colr=highpal[reg[7]&63];

    for (y=0;y<224;y++)
    {
      unsigned char *pstr;
      int pixst,pixcnt;
      pixst=8; pixcnt=8+4*8-8;

      pstr=lne+depth_to_bytes(bmp_depth)*pixst;
      if (bmp_depth<=8)       clear_to_col_8(pstr,pixcnt,colr);
      else if (bmp_depth<=16) clear_to_col_16(pstr,pixcnt,colr);
      else if (bmp_depth<=24) clear_to_col_24(pstr,pixcnt,colr);
      else if (bmp_depth<=32) clear_to_col_32(pstr,pixcnt,colr);

      pixst=8+36*8; pixcnt=8+320 - (8+36*8);
      pstr=lne+depth_to_bytes(bmp_depth)*pixst;
      if (bmp_depth<=8)       clear_to_col_8(pstr,pixcnt,colr);
      else if (bmp_depth<=16) clear_to_col_16(pstr,pixcnt,colr);
      else if (bmp_depth<=24) clear_to_col_24(pstr,pixcnt,colr);
      else if (bmp_depth<=32) clear_to_col_32(pstr,pixcnt,colr);

//    rectfill(bmp,8+0,8+0,8+4*8-1,8+224-1,highpal[reg[7]&63]);
//    rectfill(bmp,8+36*8,8+0,8+320-1,8+224-1,highpal[reg[7]&63]);
      lne+=lpitch;
    }
  }
#endif

  if (0)
  {
    int x,y;
    // show tiles
    for (y=0;y<32;y++)
      for (x=0;x<16;x++)
        draw_tile(x*8,y*8,(0xc000/32)+y*16+x);
  }

  
  if (0)
  {
    int x,y;
    // show cols
    for (y=0;y<4;y++)
      for (x=0;x<16;x++)
;//        rectfill(bmp,x*4,y*4,x*4+3,y*4+3,highpal[(y*16+x)&63]);
  }

  if (0)
  {
    int i;
    // vdp registers
    for (i=0;i<24;i++)
    {
//      int col=255;
      sprintf (tmp,"[%.2x]=%.2x\n",i,reg[i]);
//      text_mode(-1);
//      textout(bmp,font,tmp,1,i*8+1,254);
//      textout(bmp,font,tmp,0,i*8,col);
    }
  }

  
  return 0;
}
