// DGen v1.05+
// Megadrive C++ module - misc memory

#include <stdio.h>
#include "md.h"
// md.h also has include fm.h

// This is to handle memory accesses outside of ROM and RAM
// e.g. c00000 and a00000 addresses
// REMEMBER NOT TO USE ANY STATIC variables, because they
// will exist thoughout ALL megadrives!

extern FILE *debug_log;

inline unsigned short md::calculate_coo8()
{
  // c008 goes from 0000 to ff7f (i think) in one frame
  int x=0,line_len;
#ifdef COMPILE_WITH_STAR
  if (cpu_emu==0) x=s68000readOdometer()-odo_line_start;
#endif
#ifdef COMPILE_WITH_MUSA
  if (cpu_emu==1)
  x=odo-odo_line_start;
#endif

  x<<=7;
  line_len=odo_line_end-odo_line_start;
  if (line_len>0) x/=line_len;
  if (x<0) x=0; if (x>0x7f) x=0x7f;
  return (ras<<8)+x;

/*
  int scale;
  // Yuck - had to use double... FIX THIS PLEASE!

#ifdef COMPILE_WITH_STAR
  if (cpu_emu==0)
  scale=(int) (((double)s68000readOdometer()*0x10000)/odo_frame_max);
#endif
  if (cpu_emu==1)
    scale=(int) (((double)odo*0x10000)/odo_frame_max);

  if (scale>0xffff) scale=0xffff;
  if (scale<0) scale=0;

  scale=(scale&0xff00)|((scale&0xff)>>1);
  return scale;
*/
}


unsigned char md::z80_read(unsigned int a)
{
  if ((a&0xe000)==0x4000) // 00 01 02 03 - not sure about reads
  {
    return myfm_read(a&3);
  }
  if (a>=0x8000)
  {
    int addr68k;
    addr68k=z80_bank68k<<15;
    addr68k+=a&0x7fff;
    return misc_readbyte(addr68k);
  }

  return z80ram[a&0xffff];
}

unsigned short md::z80_port_read(unsigned short a)
{
  int ret=0;

  if (debug_log!=NULL)
    fprintf (debug_log,"z80 port read %.2x return %.4x\n",a,ret);
  return ret;
}
void md::z80_port_write(unsigned short a,unsigned char v)
{
  if (debug_log!=NULL)
    fprintf (debug_log,"z80 port write %.2x %.2x\n",a,v);
}


void md::z80_write(unsigned int a,unsigned char v)
{
  if ((a&0xe000)==0x4000) // 00 01 02 03
  {
    myfm_write(a&3,v);
    return ;
  }
  if (a==0x7f11)
  { mysn_write(v); return; }

  if (a==0x6000)
  {
    z80_bank68k>>=1;
    z80_bank68k+=(v&1)<<8;
    z80_bank68k&=0x1ff; // 9 bits and filled in the new top one
    return ;
  }


  if (a>=0x8000)
  {
    int addr68k;
    addr68k=z80_bank68k<<15;
    addr68k+=a&0x7fff;
    misc_writebyte(addr68k,v);
  }
  z80ram[a&0xffff]=v;
}

unsigned md::misc_readbyte(unsigned a)
{
  a&=0x00ffffff; // in case stars didn't clip to 24-bit bus

  // In case it's a rom read after all
  if (a<0xa00000)
    if ((signed)(a^1)<romlen) { return rom[(a^1)]; }

  if (a>=0xc00012) if (a<=0xfeffff)
  {
#ifdef DEBUG_ON
    if (debug_log) fprintf (debug_log,"read  %6x\n",a);
#endif
  }

  // In case it's a ram read after all
  if (a>=0xe00000) if (a<=0xffffff) { return ram[(a^1)&0xffff]; }

#ifdef DEBUG_ON
  if ((a&0xfffffffc)==0x00c00000) // 00-03
    if (debug_log)
      fprintf(debug_log,"odo=%9d : readb (%.6x)\n",s68000readOdometer(),a);
#endif

  // GFX data read (byte)
  if ((a&0xfffffffe)==0x00c00000) // 00 or 01
    return vdp.readbyte();


  if ((a&0xfffffffc)==0xa04000)
  {
    return myfm_read(a&3);
  }

  if ((a>=0xa00000)&&(a<0xa08000))
  {
    // Read from z80's memory
    return z80_read(a&0x7fff);
  }

  // I/O port access
  if ((a&0xfffffe)==0xa10002)
  {
    int ret=0;
    if (aoo3_six==3)
    {
      // Extended pad info
      if (aoo3_toggle==0) ret=((pad[0]>>8)&0x30)+0x00;
      else ret=((pad[0])&0x30)+0x40+((pad[0]>>16)&0xf);
    }
    else
    {
      if (aoo3_toggle==0)
      {
        if (aoo3_six==4) ret=((pad[0]>>8)&0x30)+0x00+0x0f;
        else             ret=((pad[0]>>8)&0x30)+0x00+(pad[0]&3);
      }
      else ret=((pad[0])&0x30)+0x40+(pad[0]&15);
    }
//    if (debug_log) fprintf (debug_log,"%.7x ==%.2x six3 = %d six5 = %d odo=%d\n",a,ret,aoo3_six,aoo5_six,s68000readOdometer());
    return ret;
  }

  if ((a&0xfffffe)==0xa10004)
  {
    int ret=0;
    if (aoo5_six==3)
    {
      // Extended pad info
      if (aoo5_toggle==0x00) ret=((pad[1]>>8)&0x30)+0x00;
      else ret=((pad[1])&0x30)+0x40+((pad[1]>>16)&0xf);
    }
    else
    {
      if (aoo5_toggle==0x00)
      {
        if (aoo5_six==4) ret=((pad[1]>>8)&0x30)+0x00+0x0f;
        else             ret=((pad[1]>>8)&0x30)+0x00+(pad[1]&3);
      }
      else ret=((pad[1])&0x30)+0x40+(pad[1]&15);
    }
    return ret;
  }


  if (a==0xa10001)
  {
    // overseas/pal/disk/0/ md version (0-f) - may make games act different!
    if ((country_ver&0xff0)==0xff0)
    {
      // autodetect country
      int headcode=0x80,avail=0;
      int i;
      for (i=0;i<3;i++)
      {
        int ch=misc_readbyte(0x1f0+i);
             if (ch=='U') avail|=1;
        else if (ch=='E') avail|=2;
        else if (ch=='J') avail|=4;
      }
           if (avail&1) headcode=0x80;
      else if (avail&2) headcode=0xc0;
      else if (avail&4) headcode=0x00;
      return headcode+(country_ver&0x0f);
    }
    else return country_ver;
  }

  if (a==0xa11000) return 0xff;
  if (a==0xa11001) return 0xff;
  if (a==0xa11100)
  {
    return z80_online;
  }
  if (a==0xa11101) return 0x00;
  if (a==0xa11200) return 0xff;
  if (a==0xa11201) return 0xff;

  // Genecyst style - toggles fifo full/empty
  if (a==0xc00004)
  {
    unsigned char tmp=coo4;
    coo4^=0x03;
    return tmp;
  }
  if (a==0xc00005)
  {
    return coo5;
// Think this is handled by our hints...
//    unsigned char tmp=coo5;
//    if ((coo5&8)==0) coo5^=0x04;
//    return tmp;
  }

  if (a==0xc00008)
    return (calculate_coo8()>>8)&0xff;
  if (a==0xc00009)
    return calculate_coo8()&0xff;

  return 0x00;

}

void md::misc_writebyte(unsigned a,unsigned d)
{
  a&=0x00ffffff; // in case stars didn't clip to 24-bit bus

#ifdef DEBUG_ON
  if (a>=0xc00012) if (a<=0xfeffff)
  {
    if (debug_log) fprintf (debug_log,"write %6x %2x\n",a,d);
  }
#endif
  // In case it's a ram write after all

  if (a>=0xe00000) if (a<=0xffffff) { ram[(a^1)&0xffff]=d; return; }

#ifdef DEBUG_ON
  if ((a&0xfffffff8)==0x00c00000) // 00-07
    if (debug_log)
      fprintf(debug_log,"odo=%9d : writeb(%.6x,%.2x)\n",s68000readOdometer(),a,d);
#endif

  // GFX data write (byte)
  if ((a&0xfffffffe)==0x00c00000) // 00 or 01
    { vdp.writebyte(d); return; }

  if ((a&0xfffffffc)==0xa04000)
  {
    myfm_write(a&3,d);
    return;
  }

  if (a==0xc00011)
  {
    mysn_write(d); return;
  }

  if (a==0xa11100)
  {
//d8 (W) 0: BUSREQ CANCEL
//       1: BUSREQ REQUEST
    if (d==1) z80_online=0;
    else z80_online=1;
    return ;
  }
  if (a==0xa11101) return;
  if (a==0xa11200)
  {
    if (d==0) mz80reset(); return ;
  }
  if (a==0xa11201) return;

  if ((a>=0xa00000)&&(a<0xa08000))
  {
    // Write into z80's memory
    z80_write(a&0x7fff,d);
    return;
  }

  // I/O port access
  if (a==0xa10003)
  {
    if (aoo3_six>=0 && (d&0x40)==0 && aoo3_toggle ) aoo3_six++;

    if (aoo3_six>0xc00000) aoo3_six&=~0x400000;
    // keep it circling around a high value

    if (d&0x40) aoo3_toggle=1; else aoo3_toggle=0;
    return;
  }
  if (a==0xa10005)
  {
    if (aoo5_six>=0 && (d&0x40)==0 && aoo5_toggle ) aoo5_six++;

    if (aoo5_six>0xc00000) aoo5_six&=~0x400000;
    // keep it circling around a high value

    if (d&0x40) aoo5_toggle=1; else aoo5_toggle=0;
    return;
  }

  return;
}

unsigned md::misc_readword(unsigned a)
{
  a&=0x00ffffff; // in case stars didn't clip to 24-bit bus

  if ((a&0xfffffffc)==0x00c00000) // 00 or 02
  {
    unsigned ret;
    ret=vdp.readword();
#ifdef DEBUG_ON
    if (debug_log)
      fprintf(debug_log,"odo=%9d : readw (%.6x) ret=%.4x\n",s68000readOdometer(),a,ret);
#endif
    return ret;
  }

  // else pass onto readbyte
  return ((misc_readbyte(a)<<8)+misc_readbyte(a+1));
}

void md::misc_writeword(unsigned a,unsigned d)
{
  a&=0x00ffffff; // in case stars didn't clip to 24-bit bus

#ifdef DEBUG_ON
  if ((a&0xfffffff8)==0x00c00000) // 00-07
    if (debug_log)
      fprintf(debug_log,"odo=%9d : writew(%.6x,%.4x)\n",s68000readOdometer(),a,d);
#endif

  // GFX data write (word)
  if ((a&0xfffffffc)==0x00c00000) // 00 or 02
  {
    vdp.writeword(d);
    return;
  }

  if ((a&0xfffffffc)==0x00c00004) // 04 or 06
  {
    if (coo_waiting)
    {
      // Okay completed the vdp command
      coo_cmd|=d; coo_waiting=0; 
      vdp.command(coo_cmd);
      return;
    }
    if ((d&0xc000)==0x8000) { vdp.reg[(d>>8)&0x1f]=d; return; }
    coo_cmd=d<<16; coo_waiting=1;
    return;
  }

  // else pass onto writebyte
  misc_writebyte(a,d>>8);
  misc_writebyte(a+1,d&255);
}

